import json
import time
import requests
import logging

from binance_aio.util import cleanNoneValue, encoded_string
from .authentication import hmac_hashing
from .error import ClientError, ServerError
from .__version__ import __version__


LOGGER = logging.getLogger('__name__')


class BinanceRestAuth(object):
    def __init__(
        self,
        api_key:str,
        api_secret:str
    ):
        self.base_url = 'https://api.binance.com'
        self.api_key = api_key
        self.api_secret = api_secret
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Content-Type": "application/json;charset=utf-8",
                "User-Agent": "akross/" + __version__,
                "X-MBX-APIKEY": api_key,
            }
        )

    def _get_sign(self, payload):
        return hmac_hashing(self.api_secret, payload)

    def _prepare_params(self, params):
        return encoded_string(cleanNoneValue(params))

    def _dispatch_request(self, http_method):
        return {
            "GET": self.session.get,
            "DELETE": self.session.delete,
            "PUT": self.session.put,
            "POST": self.session.post,
        }.get(http_method, "GET")

    def send_request(self, http_method, url_path, payload=None):
        if payload is None:
            payload = {}
        url = self.base_url + url_path
        LOGGER.debug("url: " + url)
        params = cleanNoneValue(
            {
                "url": url,
                "params": self._prepare_params(payload),
                # "timeout": self.timeout,
                # "proxies": self.proxies,
            }
        )
        response = self._dispatch_request(http_method)(**params)
        LOGGER.debug("raw response from server:" + response.text)
        self._handle_exception(response)

        try:
            data = response.json()
        except ValueError:
            data = response.text
        result = {}

        if len(result) != 0:
            result["data"] = data
            return result

        return data

    def sign_request(self, http_method, url_path, payload=None):
        if payload is None:
            payload = {}
        payload["timestamp"] = int(time.time() * 1000)
        query_string = self._prepare_params(payload)
        payload["signature"] = self._get_sign(query_string)
        return self.send_request(http_method, url_path, payload)

    def _handle_exception(self, response):
        status_code = response.status_code
        if status_code < 400:
            return
        if 400 <= status_code < 500:
            try:
                err = json.loads(response.text)
            except json.JSONDecodeError:
                raise ClientError(
                    status_code, None, response.text, None, response.headers
                )
            error_data = None
            if "data" in err:
                error_data = err["data"]
            raise ClientError(
                status_code, err["code"], err["msg"], response.headers, error_data
            )
        raise ServerError(status_code, response.text)

    def _get_user_asset(self, **kwargs):
        return self.sign_request('POST', '/sapi/v3/asset/getUserAsset', {**kwargs})

    def new_listen_key(self):
        # https://binance-docs.github.io/apidocs/spot/en/#listen-key-spot
        url_path = "/api/v3/userDataStream"
        return self.send_request("POST", url_path)

    def _renew_listen_key(self, listenKey: str):
        # Ping/Keep-alive a ListenKey (USER_STREAM)
        # https://binance-docs.github.io/apidocs/spot/en/#listen-key-spot
        url_path = "/api/v3/userDataStream"
        return self.send_request("PUT", url_path, {"listenKey": listenKey})

    # Margin
    def new_margin_listen_key(self):
        # https://binance-docs.github.io/apidocs/spot/en/#listen-key-margin
        url_path = "/sapi/v1/userDataStream"
        return self.send_request("POST", url_path)

    def _new_order(self, symbol: str, side: str, type: str, **kwargs):
        """New Order (TRADE)

        Post a new order

        POST /api/v3/order

        https://binance-docs.github.io/apidocs/spot/en/#new-order-trade

        Args:
            symbol (str)
            side (str) - 'BUY' or 'SELL'
            type (str) - 'LIMIT', 'MARKET', 'STOP_LOSS', 'STOP_LOSS_LIMIT', 'TAKE_PROFIT', 'TAKE_PROFIT_LIMIT', 'LIMIT_MAKER'
        Keyword Args:
            timeInForce (str, optional) - GTC(Good-till-cancel), IOC(Immediate-or-cancel), FOK
            quantity (float, optional)
            quoteOrderQty (float, optional)
            price (float, optional)
            newClientOrderId (str, optional): A unique id among open orders. Automatically generated if not sent.
            strategyId (int, optional)
            strategyType (int, optional): The value cannot be less than 1000000.
            stopPrice (float, optional): Used with STOP_LOSS, STOP_LOSS_LIMIT, TAKE_PROFIT, and TAKE_PROFIT_LIMIT orders.
            icebergQty (float, optional): Used with LIMIT, STOP_LOSS_LIMIT, and TAKE_PROFIT_LIMIT to create an iceberg order.
            newOrderRespType (str, optional): Set the response JSON. ACK, RESULT, or FULL;
                    MARKET and LIMIT order types default to FULL, all other orders default to ACK.
            recvWindow (int, optional): The value cannot be greater than 60000

            Type
            LIMIT - timeInForce, quantity, price
            MARKET - quantity or quoteOrderQty
        """
        params = {"symbol": symbol, "side": side, "type": type, **kwargs}
        url_path = "/api/v3/order"
        return self.sign_request("POST", url_path, params)


    def _cancel_order(self, symbol: str, **kwargs):
        """Cancel Order (TRADE)

        Cancel an active order.

        DELETE /api/v3/order

        https://binance-docs.github.io/apidocs/spot/en/#cancel-order-trade

        Args:
            symbol (str)
        Keyword Args:
            orderId (int, optional)
            origClientOrderId (str, optional)
            newClientOrderId (str, optional)
            recvWindow (int, optional): The value cannot be greater than 60000
        """
        url_path = "/api/v3/order"
        payload = {"symbol": symbol, **kwargs}
        return self.sign_request("DELETE", url_path, payload)

    def _get_order(self, symbol, **kwargs):
        """Query Order (USER_DATA)

        Check an order's status.

        GET /api/v3/order

        https://binance-docs.github.io/apidocs/spot/en/#query-order-user_data

        Args:
            symbol (str)
        Keyword Args:
            orderId (int, optional)
            origClientOrderId (str, optional)
            recvWindow (int, optional): The value cannot be greater than 60000
        """
        url_path = "/api/v3/order"
        payload = {"symbol": symbol, **kwargs}
        return self.sign_request("GET", url_path, payload)

    def _get_open_orders(self, symbol=None, **kwargs):
        """Current Open Orders (USER_DATA)

        Get all open orders on a symbol.

        GET /api/v3/openOrders

        https://binance-docs.github.io/apidocs/spot/en/#current-open-orders-user_data

        Args:
            symbol (str, optional)
        Keyword Args:
            recvWindow (int, optional): The value cannot be greater than 60000
        """
        url_path = "/api/v3/openOrders"
        payload = {"symbol": symbol, **kwargs}
        return self.sign_request("GET", url_path, payload)

    def _get_trade_list(self, symbol:str, **kwargs):
        """Account Trade List (USER_DATA)

        Get trades for a specific account and symbol.

        GET /api/v3/myTrades

        https://binance-docs.github.io/apidocs/spot/en/#account-trade-list-user_data

        Args:
            symbol (str)
        Keyword Args:
            fromId (int, optional): TradeId to fetch from. Default gets most recent trades.
            orderId (int, optional): This can only be used in combination with symbol
            startTime (int, optional)
            endTime (int, optional)
            limit (int, optional): Default Value: 500; Max Value: 1000
            recvWindow (int, optional): The value cannot be greater than 60000
        ex:
        [
            {
                "symbol": "BNBBTC",
                "id": 28457,
                "orderId": 100234,
                "orderListId": -1, //Unless OCO, the value will always be -1
                "price": "4.00000100",
                "qty": "12.00000000",
                "quoteQty": "48.000012",
                "commission": "10.10000000",
                "commissionAsset": "BNB",
                "time": 1499865549590,
                "isBuyer": true,
                "isMaker": false,
                "isBestMatch": true
            }
        ]
        """
        url_path = "/api/v3/myTrades"
        payload = {"symbol": symbol, **kwargs}
        return self.sign_request("GET", url_path, payload)

if __name__ == '__main__':
    import os
    LOG_FORMAT = ('%(levelname) -10s %(asctime)s %(name) -30s %(funcName) '
                '-35s %(lineno) -5d: %(message)s')
    logging.basicConfig(level=logging.DEBUG, format=LOG_FORMAT)
    ba = BinanceRestAuth(os.getenv('BINANCE_API_KEY'), os.getenv('BINANCE_API_PASSWORD'))
    # print(ba._get_user_asset())
    l = ba._get_trade_list('BTCUSDT', limit=1000)
    for trade in l:
        print(trade)
    



import asyncio
import websockets
import json
import ssl
import logging
from datetime import datetime, timedelta

LOGGER = logging.getLogger(__name__)
# logger.setLevel(logging.INFO)
# logger.addHandler(logging.StreamHandler())

ssl_context = ssl.create_default_context()
ssl_context.check_hostname = False
ssl_context.verify_mode = ssl.CERT_NONE


class BinanceWsAsync(object):
    MAX_SUBSCRIBE_COUNT = 100 # 1024

    def __init__(self):
        # multi-stream should be '/stream' suffix
        self._url = 'wss://stream.binance.com:9443/stream'
        self._ws = None
        self._response_dict = {}
        self._seq_id = 0
        self._running = False
        self._receive_task = None
        self._callbacks = {}
        self._session_start_time = datetime.now()
        self._last_message_time = datetime.now()
        self._last_messages = []

    @property
    def seq_id(self):
        self._seq_id += 1
        return self._seq_id
    
    async def _check_session_time(self):
        while True:
            await asyncio.sleep(600)
            # binance keeps 24hours for single ws session
            if datetime.now() - self._session_start_time > timedelta(hours=23):
                await self.recover()
                break

    async def _send_data(self, payload, blocking):
        if blocking:
            self._response_dict[self._seq_id] = asyncio.Future()
        packet = json.dumps(payload, ensure_ascii=False)

        diff = datetime.now() - self._last_message_time
        delay = timedelta(milliseconds=350) # 5 message per 1 sec but pong response included
        if diff < delay:
            await asyncio.sleep((delay - diff).total_seconds())
        
        if self._ws:
            await self._ws.send(str(packet)) # since binance accepts TEXT not BINARY
        self._last_message_time = datetime.now()
        if blocking:
            LOGGER.info('WAITING seq_id : %d', self._seq_id)
            return await self._response_dict[self._seq_id]
        return True

    async def recover(self):
        self._running = False

        LOGGER.warning('recover')
        if self._receive_task:
            LOGGER.warning('cancel task')
            self._receive_task.cancel()
        
        if self._ws:
            LOGGER.warning('close socket')
            asyncio.create_task(self._ws.close())
            # await self._ws.close()
        LOGGER.warning('rerunning ws')
        self._seq_id = 0
        await self.run()
        LOGGER.warning('subscribe exising list')
        for key in self._callbacks.keys():
            if key.endswith('aggTrade'):
                await self.subscribe_aggtrade(key.split('@')[0], self._callbacks[key])
            elif key.endswith('trade'):
                await self.subscribe_trade(key.split('@')[0], self._callbacks[key])
            elif key.endswith('depth20@100ms'):
                await self.subscribe_orderbook(key.split('@')[0], self._callbacks[key])
        LOGGER.warning('recover done')

    async def list_subscribe(self):
        data = {"method": "LIST_SUBSCRIPTIONS", "id": self.seq_id}
        return await self._send_data(data, True)

    async def subscribe(self, subscribe_type, symbol, callback, blocking):
        if not self._running or len(self._callbacks) > BinanceWsAsync.MAX_SUBSCRIBE_COUNT:
            LOGGER.warning('not running or subscribe count reached to max(%d/%d)',
                            len(self._callbacks),
                            BinanceWsAsync.MAX_SUBSCRIBE_COUNT)
            return False

        symbol = f'{symbol.lower()}@{subscribe_type}'
        # do not check duplicated, below will block subscribe during recover
        # if symbol in self._callbacks:
        #     LOGGER.warning('already subscribe %s, skip subscribe duplicated(maybe agent msg block timeout)', symbol)
        #     return True
        self._callbacks[symbol] = callback
        data = {"method": "SUBSCRIBE", "params": [symbol], "id": self.seq_id}
        LOGGER.info('subscribe %s, id:%d', symbol, self._seq_id)
        return await self._send_data(data, blocking)

    async def unsubscribe(self, subscribe_type, symbol, blocking):
        symbol = f'{symbol.lower()}@{subscribe_type}'
        if symbol in self._callbacks:
            del self._callbacks[symbol]
            if not self._running:
                LOGGER.warning('unsubscribe without sending data(recover) %s', symbol)
                return True
            data = {"method": "UNSUBSCRIBE", "params": [symbol], "id": self.seq_id}
            return await self._send_data(data, blocking)
        LOGGER.warning('%s is not subscribe status', symbol)
        return False        

    async def subscribe_orderbook(self, symbol, callback, blocking=True):
        return await self.subscribe('depth20@100ms', symbol, callback, blocking)

    async def unsubscribe_orderbook(self, symbol, blocking=True):
        return await self.unsubscribe('depth20@100ms', symbol, blocking)

    async def subscribe_aggtrade(self, symbol, callback, blocking=True):
        return await self.subscribe('aggTrade', symbol, callback, blocking)
    
    async def unsubscribe_aggtrade(self, symbol, blocking=True):
        return await self.unsubscribe('aggTrade', symbol, blocking)

    async def subscribe_trade(self, symbol, callback, blocking=True):
        return await self.subscribe('trade', symbol, callback, blocking)
    
    async def unsubscribe_trade(self, symbol, blocking=True):
        return await self.unsubscribe('trade', symbol, blocking)

    async def print_last_messages(self):
        for msg in self._last_messages:
            print(msg)

    async def receive_data(self):
        while self._running:
            try:
                if self._ws:
                    recv = await self._ws.recv()
                    LOGGER.debug('recv %s', recv)
                    data = json.loads(recv)
                    if 'stream' in data and 'data' in data and data['stream'] in self._callbacks:
                        await self._callbacks[data['stream']](data['data'])
                    elif 'id' in data and data['id'] in self._response_dict:
                        LOGGER.info('response ID: %d', data['id'])
                        if 'result' in data:
                            self._response_dict[data['id']].set_result(data['result'])
                        else:
                            self._response_dict[data['id']].set_result(0)
                    
                    self._last_messages.append(data)
                    if len(self._last_messages) > 10:
                        self._last_messages = self._last_messages[-10:]
            except Exception as e:
                LOGGER.error('error(try recover) %s', str(e))
                await self.recover()
                break
        LOGGER.error('stop receive ws data')

    async def run(self):
        # just respond to pong frame from server
        while True:
            try:
                LOGGER.info('try to connect %s', self._url)
                self._ws = await websockets.connect(self._url, ssl=ssl_context, ping_interval=None)
                LOGGER.info('connected')
                break
            except Exception as e:
                LOGGER.error('cannot initiate connection %s', str(e))
                await asyncio.sleep(1)

        self._session_start_time = datetime.now()
        asyncio.create_task(self._check_session_time())
        self._running = True
        self._receive_task = asyncio.create_task(self.receive_data())






import asyncio
import websockets
import json
import ssl
import logging
from datetime import datetime, timedelta

from binance_aio.binance_rest_auth import BinanceRestAuth
from binance_aio.util import cleanNoneValue, encoded_string


LOGGER = logging.getLogger(__name__)


ssl_context = ssl.create_default_context()
ssl_context.check_hostname = False
ssl_context.verify_mode = ssl.CERT_NONE

"""
{
  "e": "outboundAccountPosition", //Event type
  "E": 1564034571105,             //Event Time
  "u": 1564034571073,             //Time of last account update
  "B": [                          //Balances Array
    {
      "a": "ETH",                 //Asset
      "f": "10000.000000",        //Free
      "l": "0.000000"             //Locked
    }
  ]
}

{
  "e": "balanceUpdate",         //Event Type
  "E": 1573200697110,           //Event Time
  "a": "BTC",                   //Asset
  "d": "100.00000000",          //Balance Delta
  "T": 1573200697068            //Clear Time
}

"""


class BinanceWsUserdata(BinanceRestAuth):

    def __init__(self, api_key, api_secret):
        # multi-stream should be '/stream' suffix
        super().__init__(api_key, api_secret)
        self._url = 'wss://stream.binance.com:9443/ws/'
        self._ws = None
        self._running = False
        self._receive_task = None
        self._callbacks = {}
        self._session_start_time = datetime.now()
        self._ping_sent_time = datetime.now()
        self._listen_key = ''
    
    async def _check_session_time(self):
        while True:
            await asyncio.sleep(600)
            # binance keeps 24hours for single ws session
            if datetime.now() - self._ping_sent_time > timedelta(minutes=45):
                LOGGER.info('send ping')
                self._ping_sent_time = datetime.now()
                await self.renew_listen_key()
            elif datetime.now() - self._session_start_time > timedelta(hours=23):
                await self.recover()
                break

    async def renew_listen_key(self):
        LOGGER.info('')
        return self._renew_listen_key(self._listen_key)

    async def get_asset_list(self):
        res = self._get_user_asset(recvWindow=60000)
        return res

    async def get_open_orders(self):
        res = self._get_open_orders(recvWindow=60000)
        LOGGER.info('get open order %s', res)
        return res

    async def get_trade_list(self, symbol:str):
        res = self._get_trade_list(symbol, limit=1000)
        LOGGER.info('get trade list')
        return res

    async def query_order(self, symbol: str, orderId:int):
        res = self._get_order(symbol, orderId=orderId, recvWindow=60000)
        LOGGER.info('query order %s', res)
        return res

    async def new_order(self, symbol: str, side: str, type: str, **kwargs):
        LOGGER.info('order %s %s %s %s', symbol, side, type, kwargs)
        try:
            res = self._new_order(symbol, side, type, **kwargs)
        except Exception as e:
            if len(e.args) > 2:
                raise Exception(str(e.args[0]) + ':' + e.args[2])
            else:
                raise Exception(str(e))
        return res

    async def cancel_order(self, symbol: str, **kwargs):
        LOGGER.info('cancel order %s %s', symbol, kwargs)
        try:
            res = self._cancel_order(symbol, **kwargs)
        except Exception as e:
            if len(e.args) > 2:
                raise Exception(str(e.args[0]) + ':' + e.args[2])
            else:
                raise Exception(str(e))
        return res

    async def recover(self):
        self._running = False

        LOGGER.warning('start recover')
        if self._receive_task:
            LOGGER.warning('cancel task')
            self._receive_task.cancel()
        
        if self._ws:
            LOGGER.warning('close socket')
            asyncio.create_task(self._ws.close())
        LOGGER.warning('rerunning ws')
        
        await self.run()
        LOGGER.warning('recover done')

    def subscribe(self, callback, obj):
        self._callbacks[obj] = callback

    def unsubscribe(self, obj):
        if obj in self._callbacks:
            del self._callbacks[obj]
        
    async def receive_data(self):
        while self._running:
            recv = await self._ws.recv()
            try:
                data = json.loads(recv)
                LOGGER.info('userdata recv %s', recv)
                for callback in self._callbacks.values():
                    await callback(data)
            except Exception as e:
                LOGGER.error(str(e))
        LOGGER.error('stop receive ws data')

    async def run(self):
        # just respond to pong frame from server
        while True:
            try:
                self._listen_key = self.new_listen_key()['listenKey']
                LOGGER.info('get listen key %s', self._listen_key)
                LOGGER.info('try to connect %s', self._url + self._listen_key)
                self._ws = await websockets.connect(
                    self._url + self._listen_key, ssl=ssl_context)
                LOGGER.info('connected')
                break
            except Exception as e:
                LOGGER.error('cannot initiate connection %s', str(e))
                await asyncio.sleep(1)
        
        self._session_start_time = datetime.now()
        self._ping_sent_time = datetime.now()
        asyncio.create_task(self._check_session_time())
        self._running = True
        self._receive_task = asyncio.create_task(self.receive_data())


    





from urllib.parse import urlencode


def cleanNoneValue(d) -> dict:
    out = {}
    for k in d.keys():
        if d[k] is not None:
            out[k] = d[k]
    return out


def encoded_string(query):
    return urlencode(query, True).replace("%40", "@")


import asyncio
import websockets
import ssl

ssl_context = ssl.create_default_context()
ssl_context.check_hostname = False
ssl_context.verify_mode = ssl.CERT_NONE

async def main():
    ws = await websockets.connect('wss://stream.binance.com:9443/stream', ping_interval=None, ssl=ssl_context)



asyncio.run(main())